"""
Provide a way to enable logging by setting LOG_LEVEL environment variable
"""
import logging
import os

def setup_logging():
    # Set default logging level to WARNING if LOG_LEVEL is not set
    log_level = os.getenv("LOG_LEVEL", "WARNING").upper()
    
    # Configure logging
    logging.basicConfig(level=getattr(logging, log_level, logging.WARNING))


import re
from dataclasses import dataclass
import numpy as np

@dataclass
class Segment:
    text: str
    next_pause: float
    
    def create_pause(self, sample_rate: int):
        """
        Return array of zeros (silence) with size of the next_pause
        """
        num_samples = int(self.next_pause * sample_rate)
        return np.zeros(num_samples, dtype=np.float32)

class SegmentExtractor:
    def __init__(self, default_pause: float = 0.02, question_pause: float = 0.05, period_pause: float = 0.05, new_line_pause = 0.3):
        self.default_pause = default_pause
        self.question_pause = question_pause
        self.period_pause = period_pause
        self.new_line_pause = new_line_pause
    
    def extract_segments(self, text: str):
        """
        Break the text into segment items. eg. on '?' or '.' or '!' or '\n'
        Used later to add pauses when speaking
        """
        sentences = re.split(r'([.?!:\n])', text)
        for i in range(0, len(sentences) - 1, 2):
            sentence = sentences[i].strip()
            punctuation = sentences[i + 1]
            if sentence:  # Ensure the sentence is not empty
                if punctuation == '.':
                    yield Segment(text=f"{sentence}{punctuation}", next_pause=self.period_pause)
                elif punctuation == '?':
                    yield Segment(text=f"{sentence}{punctuation}", next_pause=self.question_pause)
                elif punctuation == '\n':
                    yield Segment(text=f"{sentence}{punctuation}.", next_pause=self.new_line_pause)
                else:
                    yield Segment(text=f"{sentence}{punctuation}", next_pause=self.default_pause)
        last_sentence = sentences[-1].strip()
        if last_sentence:
            yield Segment(text=f"{last_sentence}.", next_pause=self.default_pause)


from israwave.tensors import FloatArray
from .model import Model
import soundfile as sf
import sounddevice as sd
import os
from israwave.logging import setup_logging
from pathlib import Path

setup_logging()

class WaveForm:
    def __init__(self, samples: FloatArray, sample_rate: int) -> None:
        self.samples =  samples
        self.sample_rate = sample_rate
    
    def save(self, path: str):
        if os.path.exists(path):
            os.remove(path)
        sf.write(path, self.samples, self.sample_rate)
        
    def play(self):
        sd.play(self.samples, self.sample_rate)
        sd.wait()

class IsraWave:
    def __init__(self, model_path: str, espeak_data_path: str) -> None:
        # Check if the speech model exists
        if not Path(model_path).exists():
            raise FileNotFoundError(
                f"Speech model not found at {model_path}\n"
                "Please download and prepare the model using the following commands:\n"
                "wget https://github.com/thewh1teagle/israwave/releases/download/v0.1.0/israwave.onnx"
            )

        # Check if the espeak data folder exists
        if not Path(espeak_data_path).exists():
            raise FileNotFoundError(
                f"Espeak data folder not found at {espeak_data_path}\n"
                "Please download and extract the espeak data using the following commands:\n"
                "wget https://github.com/thewh1teagle/israwave/releases/download/v0.1.0/espeak-ng-data.tar.gz\n"
                "tar xf espeak-ng-data.tar.gz"
            )

        self.model = Model(model_path, espeak_data_path)
        self.sample_rate = self.model.sample_rate
    
    def create(self, text, rate = 1.0, pitch = 1.0, energy = 1.0):
        """create speech waveform

        Args:
            text str: _description_
            rate (float, optional): Control rate. Defaults to 1.0.
            pitch (float, optional): Control pitch. Defaults to 1.0.
            energy (float, optional): Control energy. Defaults to 1.0.

        Returns:
            _type_: _description_
        """
        inputs = self.model.prepare_input(
            text,
            d_factor=rate,
            p_factor=pitch,
            e_factor=energy,
            lang='he'
        )
        outputs = self.model.synthesise(inputs)
        wav = outputs.unbatched_wavs()[0]
        waveform = WaveForm(wav, self.model.sample_rate)
        return waveform
        
        


from pathlib import Path
import json

CONFIG_PATH = Path(__file__).parent / 'config.json'
with open(CONFIG_PATH, encoding='utf-8') as f:
    CONFIG = json.load(f)
    SYMBOLS = CONFIG['symbols']
   
# Special symbols
PAD = CONFIG['PAD']
BOS = CONFIG['BOS']
EOS = CONFIG['EOS']

# Special symbol ids
PAD_ID = SYMBOLS.index(PAD)
BOS_ID = SYMBOLS.index(BOS)
EOS_ID = SYMBOLS.index(EOS)
SPACE_ID = SYMBOLS.index(" ")

# Mappings from symbol to numeric ID and vice versa:
SYMBOL_TO_ID = {s: i for i, s in enumerate(SYMBOLS)}
ID_TO_SYMBOL = {i: s for i, s in enumerate(SYMBOLS)}  # pylint: disable=unnecessary-comprehension


def phonemes_to_ids(text):
    """Converts a string of text to a sequence of IDs corresponding to the symbols in the text.
    Args:
      text: string to convert to a sequence
    Returns:
      List of integers corresponding to the symbols in the text
    """
    sequence = []
    for symbol in text:
        symbol_id = SYMBOL_TO_ID[symbol]
        sequence.append(symbol_id)
    return sequence


def ids_to_phonemes(sequence):
    """Converts a sequence of IDs back to a string"""
    result = ""
    for symbol_id in sequence:
        s = ID_TO_SYMBOL[symbol_id]
        result += s
    return result



import re
from .symbols import phonemes_to_ids
import logging

WHITESPACE_RE = re.compile(r"\s+")

class IPATokenizer:
    def __init__(self, espeak_data_path = None) -> None:
        self.espeak_data_path = espeak_data_path
    
    def preprocess_text(self, text, _language):
        return self.collapse_whitespace(text)
    
    def collapse_whitespace(self, text):
        text = re.sub(WHITESPACE_RE, " ", text)
        return text
    
    def phonemize_text(self, text: str, language: str) -> str:
        try:
            from piper_phonemize import phonemize_espeak
        except ImportError:
            raise ImportError(
                "piper-phonemize package is needed for the IPA tokenizer.\n"
                "pip install piper-phonemize\n"
                "or build it yourself from the following repository:\n"
                "https://github.com/rhasspy/piper-phonemize"
            )

        # Preprocess
        text = self.preprocess_text(text, language)
        # Phonemize        
        phonemes = phonemize_espeak(text, language, data_path=self.espeak_data_path)
        logging.debug(f"phonemes: {phonemes} text: {text}")
        return phonemes, text
    
    def tokenize(self, text, language):
        phonemes, normalized_text = self.phonemize_text(text, language)
        phonemes = [phoneme for sentence_phonemes in phonemes for phoneme in sentence_phonemes]
        phonemes = list(self.collapse_whitespace("".join(phonemes)))
        phoneme_ids = phonemes_to_ids(phonemes)
        return phoneme_ids, normalized_text


import json
import logging
from time import perf_counter
import numpy as np
import onnxruntime
from israwave.tokenizer import IPATokenizer
from .tensors import InferenceInputs, InferenceOutputs

log = logging.getLogger("infer")

class Model:

    def __init__(self, onnx_path: str, espeak_data_path: str, onnx_providers: list[str] = ["CPUExecutionProvider"]):
        session = onnxruntime.InferenceSession(onnx_path, providers=onnx_providers)
        meta = session.get_modelmeta()
        infer_params = json.loads(meta.custom_metadata_map["inference"])
        self.tokenizer=IPATokenizer(espeak_data_path)
        self.session=session
        self.name=infer_params["name"]
        self.sample_rate=infer_params["sample_rate"]
        self.inference_args=infer_params["inference_args"]
        self.speakers=infer_params["speakers"]
        self.languages=infer_params["languages"]

    def __post_init__(self):
        self.is_multispeaker = len(self.speakers) > 1
        self.is_multilanguage = len(self.languages) > 1

    def prepare_input(
        self,
        text: str,
        lang: str | None = None,
        speaker: str | int | None = None,
        d_factor: float|None=None,
        p_factor: float|None=None,
        e_factor: float|None=None,
    ) -> InferenceInputs:
        sid = None
        lid = None
        phids, clean_text = self.tokenizer.tokenize(text=text, language=lang)
        
        phids = [phids]
        input_ids = []
        lengths = []
        for phid in phids:
            input_ids.append(phid)
            lengths.append(len(phid))
        sids = [sid] * len(input_ids) if sid is not None else None
        lids = [lid] * len(input_ids) if lid is not None else None
        return InferenceInputs.from_ids_and_lengths(
            ids=input_ids,
            lengths=lengths,
            clean_text=clean_text,
            sids=sids,
            lids=lids,
            d_factor=d_factor or self.inference_args["d_factor"],
            p_factor=p_factor or self.inference_args["p_factor"],
            e_factor=e_factor or self.inference_args["e_factor"],
        )

    def synthesise(self, inference_inputs: InferenceInputs) -> InferenceOutputs:
        inference_inputs = inference_inputs.as_numpy()
        synth_outs = self.synthesise_with_values(
            x=inference_inputs.x,
            x_lengths=inference_inputs.x_lengths,
            sids=inference_inputs.sids,
            lids=inference_inputs.lids,
            d_factor=inference_inputs.d_factor,
            p_factor=inference_inputs.p_factor,
            e_factor=inference_inputs.e_factor
        )
        return InferenceOutputs(
            wav=synth_outs["wav"],
            wav_lengths=synth_outs["wav_lengths"],
            latency=synth_outs["latency"],
            rtf=synth_outs["rtf"],
        )

    def synthesise_with_values(self, x, x_lengths, sids, lids, d_factor, p_factor, e_factor):
        inputs = dict(
            x=x,
            x_lengths=x_lengths,
            scales=np.array([d_factor, p_factor, e_factor], dtype=np.float32),
        )
        t0 = perf_counter()
        wav, wav_lengths, durations = self.session.run(None, inputs)
        t_infer = perf_counter() - t0
        t_audio = wav_lengths.sum() / self.sample_rate
        rtf = t_infer / t_audio
        latency = t_infer * 1000
        return dict(wav=wav, wav_lengths=wav_lengths, rtf=rtf, latency=latency)





import dataclasses
from dataclasses import dataclass
from typing import TypeAlias
import numpy as np

FloatArray: TypeAlias = np.ndarray[np.float32]
IntArray: TypeAlias = np.ndarray[np.int64]


@dataclass
class BaseTensorContainer:

    def as_tuple(self):
        return dataclasses.astuple(self)

    def as_dict(self):
        return dataclasses.asdict(self)

    def as_numpy(self):
        data = self.as_dict()
        kwargs = {}
        for name, value in self.as_dict().items():
            if isinstance(value, np.ndarray):
                kwargs[name] = np.asarray(value)
            else:
                kwargs[name] = value
        cls = type(self)
        return cls(**kwargs)


@dataclass(kw_only=True)
class InferenceInputs(BaseTensorContainer):
    clean_text: str
    x: IntArray
    x_lengths: IntArray
    sids: IntArray|None = None
    lids: IntArray|None = None
    d_factor: float = 1.0
    p_factor: float = 1.0
    e_factor: float = 1.0

    @classmethod
    def from_ids_and_lengths(cls, ids: list[int], lengths: list[int], **kwargs) -> "Self":
        x = numpy_pad_sequences(ids).astype(np.int64)
        x_lengths = np.array(lengths, dtype=np.int64)
        instance = cls(x=x, x_lengths=x_lengths, **kwargs)
        return instance.as_numpy()


@dataclass(kw_only=True)
class InferenceOutputs(BaseTensorContainer):
    wav: FloatArray
    wav_lengths: FloatArray
    latency: int
    rtf: float
    durations: FloatArray|None = None
    pitch: FloatArray|None = None
    energy: FloatArray|None = None
    am_rtf: float|None = None
    v_rtf: float|None = None

    def __iter__(self):
        return iter(self.unbatched_wavs())

    def unbatched_wavs(self) -> list[FloatArray]:
        if isinstance(self.wav, np.ndarray):
            return numpy_unpad_sequences(self.wav, self.wav_lengths)
        else:
            raise RuntimeError("Unsupported operation")


def numpy_pad_sequences(sequences, maxlen=None, value=0):
    """Pads a list of sequences to the same length using broadcasting.

    Args:
      sequences: A list of Python lists with variable lengths.
      maxlen: The maximum length to pad the sequences to. If not specified,
        the maximum length of all sequences in the list will be used.
      value: The value to use for padding (default 0).

    Returns:
      A numpy array with shape [batch_size, maxlen] where the sequences are padded
      with the specified value.
    """

    # Get the maximum length if not specified
    if maxlen is None:
        maxlen = max(len(seq) for seq in sequences)

    # Create a numpy array with the specified value and broadcast
    padded_seqs = np.full((len(sequences), maxlen), value)
    for i, seq in enumerate(sequences):
        padded_seqs[i, : len(seq)] = seq

    return padded_seqs


def numpy_unpad_sequences(sequences, lengths):
    """Unpads a list of sequences based on a list of lengths.

    Args:
      sequences: A numpy array with shape [batch_size, feature_dim, max_len].
      lengths: A numpy array with shape [batch_size] representing the lengths
        of each sequence in the batch.

    Returns:
      A list of unpadded sequences. The i-th element of the list corresponds
      to the i-th sequence in the batch. Each sequence is a numpy array with
      variable length.
    """

    # Check if lengths argument is a list or 1D numpy array
    if not isinstance(lengths, np.ndarray) or len(lengths.shape) != 1:
        raise ValueError("lengths must be a 1D numpy array")

    # Check if sequence lengths are within bounds
    if np.any(lengths < 0) or np.any(lengths > sequences.shape[-1]):
        raise ValueError("lengths must be between 0 and max_len")

    # Get the batch size
    batch_size = sequences.shape[0]

    # Extract unpadded sequences
    unpadded_seqs = []
    for i in range(batch_size):
        unpadded_seqs.append(sequences[i, : lengths[i]])

    return unpadded_seqs

